/*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
* http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

'use strict';

const rewire = require('rewire');
const ClientOrchestratorRewire = rewire('../../lib/client/client-orchestrator');

const chai = require('chai');
chai.should();
const sinon = require('sinon');

describe('client orchestrator implementation', () => {


    describe('#init', () => {

        it('should read the number of test clients if present in the config file', () => {
            const FakeParseYaml = sinon.stub().returns({ test: { clients: {number: 7}}});
            ClientOrchestratorRewire.__set__('util.parseYaml', FakeParseYaml);

            const myOrchestrator = new ClientOrchestratorRewire();
            myOrchestrator.init();
            myOrchestrator.number.should.equal(7);

        });

        it('should default to one client in the test if not specified in the config file ', () => {
            const FakeParseYaml = sinon.stub().returns({ test: { clients: {notNumber: 2}}});
            ClientOrchestratorRewire.__set__('util.parseYaml', FakeParseYaml);

            const myOrchestrator = new ClientOrchestratorRewire();
            myOrchestrator.init();
            myOrchestrator.number.should.equal(1);
        });
    });

    describe('#startTest', () => {

        const FakeParseYaml = sinon.stub().returns({ test: { clients: {number: 7}}});
        ClientOrchestratorRewire.__set__('util.parseYaml', FakeParseYaml);
        const myOrchestrator = new ClientOrchestratorRewire();
        myOrchestrator.init();
        let _startTestStub;
        let formatResultsStub;

        beforeEach(() => {
            _startTestStub = sinon.stub();
            formatResultsStub = sinon.stub().returns('formatted');
            myOrchestrator._startTest = _startTestStub;
            myOrchestrator.formatResults = formatResultsStub;
        });

        it('should increment the updates.id variable', async () => {
            myOrchestrator.updates.id = 41;
            const testMsg = {msg: 'test msg'};
            const clientArgs = [1,2,3];
            const factory = {};

            await  myOrchestrator.startTest(testMsg, clientArgs, factory);

            myOrchestrator.updates.id.should.equal(42);
        });

        it('should call _startTest with known arguments', async () => {
            myOrchestrator.updates.id = 0;
            const testMsg = {msg: 'test msg'};
            const clientArgs = [1,2,3];
            const factory = {factory: 'a factory'};

            await  myOrchestrator.startTest(testMsg, clientArgs, factory);

            sinon.assert.calledOnce(_startTestStub);
            sinon.assert.calledWith(_startTestStub, 7, testMsg, clientArgs, [], [], factory);
        });

        it('should call formatResults', async() => {
            myOrchestrator.updates.id = 0;
            const testMsg = {msg: 'test msg'};
            const clientArgs = [1,2,3];
            const factory = {};

            await  myOrchestrator.startTest(testMsg, clientArgs, factory);

            sinon.assert.calledOnce(formatResultsStub);
        });

        it('should return the response from formatResults', async () => {
            myOrchestrator.updates.id = 0;
            const testMsg = {msg: 'test msg'};
            const clientArgs = [1,2,3];
            const factory = {};

            const response = await  myOrchestrator.startTest(testMsg, clientArgs, factory);
            response.should.equal('formatted');
        });

    });


    describe('#getUpdates', () => {

        const FakeParseYaml = sinon.stub().returns({ test: { clients: {number: 7}}});
        ClientOrchestratorRewire.__set__('util.parseYaml', FakeParseYaml);
        const myOrchestrator = new ClientOrchestratorRewire();

        it('should return the updates', () => {
            const checkVal = 'this is my update';
            // overwrite with known value
            myOrchestrator.updates = checkVal;
            // assert repsonse
            myOrchestrator.getUpdates().should.equal(checkVal);
        });

    });

    describe('#formatResults', () => {

        const FakeParseYaml = sinon.stub().returns({ test: { clients: {number: 7}}});
        ClientOrchestratorRewire.__set__('util.parseYaml', FakeParseYaml);
        const myOrchestrator = new ClientOrchestratorRewire();

        it('should group all client results into an array under a results label', () => {
            const result0 = {results: [1] , start: new Date(2018, 11, 24, 10, 33), end: new Date(2018, 11, 24, 11, 33)};
            const result1 = {results: [2] , start: new Date(2018, 11, 24, 10, 34), end: new Date(2018, 11, 24, 11, 23)};
            const result2 = {results: [3] , start: new Date(2018, 11, 24, 10, 35), end: new Date(2018, 11, 24, 11, 13)};
            const testData = [result0, result1, result2];

            const output = myOrchestrator.formatResults(testData);
            output.results.should.deep.equal([1,2,3]);
        });

        it('should determine and persist the time when all clients have started', () => {
            const compareStart = new Date(2018, 11, 24, 10, 35);
            const result0 = {results: [1] , start: new Date(2018, 11, 24, 10, 33), end: new Date(2018, 11, 24, 11, 33)};
            const result1 = {results: [2] , start: new Date(2018, 11, 24, 10, 34), end: new Date(2018, 11, 24, 11, 13)};
            const result2 = {results: [3] , start: compareStart, end: new Date(2018, 11, 24, 11, 23)};
            const testData = [result0, result1, result2];

            const output = myOrchestrator.formatResults(testData);
            output.start.should.equal(compareStart);
        });

        it('should determine and persist the last time when all clients were running', () => {
            const compareEnd = new Date(2018, 11, 24, 11, 13);
            const result0 = {results: [1] , start: new Date(2018, 11, 24, 10, 33), end: new Date(2018, 11, 24, 11, 33)};
            const result1 = {results: [2] , start: new Date(2018, 11, 24, 10, 34), end: compareEnd};
            const result2 = {results: [3] , start:  new Date(2018, 11, 24, 10, 35), end: new Date(2018, 11, 24, 11, 23)};
            const testData = [result0, result1, result2];

            const output = myOrchestrator.formatResults(testData);
            output.end.should.equal(compareEnd);
        });
    });

});
